module Main where

import Parser
import PropPretty
import Behaviour   (Trace(..), cut)
import Interpreter (interp)
import Compiler    (compile)
import Machine     (exec)
import Syntax      (Command(..), Expr(..), Name)
import Value       (Value(..), Op1(..), Op2(..))
import Pretty      ()

import Control.Monad       (liftM, liftM2, liftM3)
import Control.Monad.State (StateT)
import Data.Map            (Map)
import qualified Data.Map as Map
import Test.QuickCheck 

-- test data generators

elements' s [] = error $ "elements used with empty list in " ++ s
elements' _ xs = elements xs

arbName :: Gen Name
arbName = sized $ \n -> elements [ [c] | c <- take (n `div` 3 + 1)
                                                   ['a'..'z'] ]

instance Arbitrary Value where
  arbitrary =
    frequency [ (4, Num `liftM` arbitrary)
              , (4, Bol `liftM` arbitrary)
              , (1, return Wrong)
              ]

  shrink (Num n) = [ Num n' | n' <- shrink n ]
  shrink (Bol b) = Num 0 : [ Bol b' | b' <- shrink b ]
  shrink Wrong   = [ Num 0 ]

instance Arbitrary Expr where
  -- includes "badly typed" expressions
  arbitrary = sized arbExpr
    where
      arbExpr n =
        frequency $
          [ (4, liftM Var arbName)
          , (1, liftM Val arbitrary)
          ] ++
          concat
          [ [ (2, liftM2 Uno arbitrary arbExpr')
            , (4, liftM3 Duo arbitrary arbExpr2 arbExpr2)
            ]
          | n > 0  -- useful list comprehension trick
          ]
        where
          arbExpr' = arbExpr (n-1)
          arbExpr2 = arbExpr (n `div` 2)

  shrink (Uno o e)        = e : [ Uno o e' | e' <- shrink e ]
  shrink (Duo o e1 e2)    = e1 : e2 :
                            [ Duo o e1' e2 | e1' <- shrink e1 ] ++
                            [ Duo o e1 e2' | e2' <- shrink e2 ]
  shrink (Val v) = [ Val v' | v' <- shrink v ]
  shrink (Var x) = [ Val (Num 0), Val (Bol False), Val (Bol True) ]

instance Arbitrary Op1 where
  arbitrary = elements [ Not, Minus ]

instance Arbitrary Command where
  arbitrary = sized arbCommand
    where
      arbCommand n =
        frequency $
          [ (1, return Skip)
          , (6, liftM2 (:=) arbName arbitrary)
          , (3, liftM  Print arbitrary)
          ] ++
          concat
          [ [ (4, liftM2 (:->) arbCommand2 arbCommand2)
            , (4, liftM3 If    arbitrary arbCommand2 arbCommand2)
            , (4, liftM2 While arbitrary arbCommand')
            ]
          | n > 0  -- useful list comprehension trick
          ]
        where
          arbCommand' = arbCommand (n-1)
          arbCommand2 = arbCommand (n `div` 2)

  shrink (x := e)     = Skip : [ x := e' | e' <- shrink e ]
  shrink (c1 :-> c2)  = c1 : c2 :
                        [ c1' :-> c2 | c1' <- shrink c1 ] ++
                        [ c1 :-> c2' | c2' <- shrink c2 ]
  shrink (If e c1 c2) = c1 : c2 :
                        [ If e' c1 c2 | e' <- shrink e ] ++
                        [ If e c1' c2 | c1' <- shrink c1 ] ++
                        [ If e c1 c2' | c2' <- shrink c2 ]
  shrink (While e c)  = c :
                        [ While e' c | e' <- shrink e ] ++
                        [ While e c' | c' <- shrink c ]
  shrink (Print e)    = Skip : [ Print e' | e' <- shrink e ]
  shrink _            = []

instance Arbitrary Op2 where
  arbitrary = elements [ minBound .. maxBound ]


--
-- Generating well-typed programs

data Type = TInt | TBool | Undef
  deriving (Eq, Show)

instance Arbitrary Type where
  arbitrary = elements [TInt, TBool]

type Env = Map Name Type

type G = StateT Env Gen

unionEnv :: Env -> Env -> Env
unionEnv env1 env2 = Map.unionWith unionType env1 env2
  where unionType Undef _  =  Undef
        unionType _ Undef  =  Undef
        unionType s     t  =  s

genValue :: Type -> Gen Value
genValue TInt  = liftM Num arbitrary
genValue TBool = liftM Bol arbitrary
genValue Undef = return Wrong

genExpr :: Env -> Type -> Gen Expr
genExpr env t = sized $ \n -> gen n t
  where
    gen n t = do
      let okVars = [ x | (x, t') <- Map.toList env, t == t' ]
      frequency $
        [ (1, liftM Var $ elements okVars) | not $ null okVars ] ++
        [ (1, liftM Val $ genValue t) ] ++
        concat
        [ [ (2, uno)
          , (4, duo)
          ]
        | n > 0
        ]
     where

      gen' = gen (n-1)
      gen2 = gen (n `div` 2)

      uno = do
        let okOps = [ args | (args, resT) <- unoOps
                           , resT == t
                    ]
        (op, argT) <- elements okOps
        liftM (Uno op) (gen' argT)

      duo = do
        let okOps = [ args
                    | (args, resT) <- duoOps
                    , resT == t
                    ]
        (op, argT1, argT2) <- elements okOps
        liftM2 (Duo op) (gen2 argT1) (gen2 argT2)

unoOps :: [((Op1, Type), Type)]
unoOps = [ ((Not, TBool), TBool)
         , ((Minus, TInt), TInt)
         ]

duoOps :: [((Op2, Type, Type), Type)]
duoOps =
  [ ((And,    TBool, TBool), TBool)
  , ((Or,     TBool, TBool), TBool)
  , ((Mul,    TInt,  TInt),  TInt)
  , ((Add,    TInt,  TInt),  TInt)
  , ((Sub,    TInt,  TInt),  TInt)
  , ((Div,    TInt,  TInt),  TInt)
  , ((Mod,    TInt,  TInt),  TInt)
  , ((Less,   TInt,  TInt),  TBool)
  , ((LessEq, TInt,  TInt),  TBool)
  , ((Eq,     TBool, TBool), TBool)
  , ((Eq,     TInt,  TInt),  TBool)
  ]

genCommand :: Env -> Gen Command
genCommand env = sized $ \n -> do
    (cmd, _) <- gen n env
    return cmd
  where
    gen n env = do
      frequency $
        [ (1, genSkip)
        , (3, genPrint)
        ] ++
        [ (6, genAssign) | not $ Map.null env ] ++
        concat
        [ [ (4, genSeq)
          , (4, genIf)
          , (4, genWhile)
          ]
        | n > 0
        ]
     where
      gen' = gen (n-1)
      gen2 = gen (n `div` 2)

      arbExpr = do
        t <- arbitrary
        genExpr env t

      genSkip = return (Skip, env)

      genPrint = do
        e <- arbExpr
        return (Print e, env)

      genAssign = do
        (x, t) <- elements $ Map.toList env
        t' <- case t of
                Undef -> arbitrary
                _     -> return t
        e <- genExpr env t'
        return (x := e, Map.insert x t' env)

      genSeq = do
        (c1, env1) <- gen2 env
        (c2, env2) <- gen2 env1
        return (c1 :-> c2, env2)

      genIf = do
        e <- genExpr env TBool
        (c1, env1) <- gen2 env
        (c2, env2) <- gen2 env
        return (If e c1 c2, unionEnv env1 env2)

      genWhile = do
        e <- genExpr env TBool
        (c, env') <- gen' env
        return (While e c, unionEnv env env')

genEnv :: Gen Env
genEnv = do
  xs  <- listOf arbName  -- may contain duplicates
  ts  <- vectorOf (length xs) arbitrary
  return $ Map.fromList $ zip xs ts -- only last binding is used in case of duplicates

genInitEnv :: Gen Env
genInitEnv = do
  xs  <- listOf arbName
  return $ Map.fromList $ zip xs $ repeat Undef

arbCommand :: Gen Command
arbCommand = genCommand =<< genInitEnv

typedShrink Skip       = []
typedShrink (Print e)  = Skip : [ Print e | e <- typedShrinkE e ]
typedShrink (x := e)   = [ x := e | e <- typedShrinkE e ]
typedShrink (a :-> b)  =
  [ a :-> b | a <- typedShrink a ] ++
  [ a :-> b | b <- typedShrink b ]
typedShrink (If c a b) = a : b :
  [ If c a b | c <- typedShrinkE c ] ++
  [ If c a b | a <- typedShrink a ] ++
  [ If c a b | b <- typedShrink b ]
typedShrink (While c a) = a :
  [ While c a | c <- typedShrinkE c ] ++
  [ While c a | a <- typedShrink a ]

typedShrinkV (Num n) = [ Num n | n <- shrink n ]
typedShrinkV (Bol b) = []
typedShrinkV Wrong   = []

typedShrinkE (Val v)      = [ Val v | v <- typedShrinkV v ]
typedShrinkE (Var x)      = []
typedShrinkE (Uno op e)   = e : [ Uno op e | e <- typedShrinkE e ]
typedShrinkE (Duo op a b) = aOrb ++
  [ Duo op a' b | a' <- typedShrinkE a ] ++
  [ Duo op a b' | b' <- typedShrinkE b ]
  where
    aOrb | op `elem` [Less, Eq, LessEq] = map (Val . Bol) [False, True]
         | otherwise                    = [a, b]

inferExpr :: Env -> Expr -> Type
inferExpr env (Var x) = case Map.lookup x env of
  Nothing -> Undef
  Just t  -> t
inferExpr env (Val v) = case v of
  Num _ -> TInt
  Bol _ -> TBool
  Wrong -> Undef
inferExpr env (Uno op e)     = inferUno op (inferExpr env e)
inferExpr env (Duo op e1 e2) = inferDuo op (inferExpr env e1)
                                           (inferExpr env e2)

inferUno op t = case lookup (op, t) unoOps of
  Nothing -> Undef
  Just t  -> t

inferDuo op t1 t2 = case lookup (op, t1, t2) duoOps of
  Nothing -> Undef
  Just t  -> t

-- helper functions

(=~=) :: (Show a, Eq a) => Trace a -> Trace a -> Property
s =~= t = forAllShrink arbitrary shrink $ \(Positive n) ->
          let s' = cut n s
              t' = cut n t
          in s' == t'
--          in whenFail (debug s t) (s' == t')
--          in  collect s' (s' == t')
  where
    debug s t = do
      putStrLn $ "lhs: " ++ show s
      putStrLn $ "rhs: " ++ show t

-- properties

prop_Congruence  = forAllShrink arbCommand shrink        $ \p ->
                     interp p =~= exec (compile p)

-- A "no shrinking" version:
prop_Congruence' = forAllShrink arbCommand shrinkNothing $ \p ->
                     interp p =~= exec (compile p)

--
canCrash p = case p of
    Skip      -> False
    Print _   -> False
    If c a b  -> canBeWrong c || any canCrash [a, b]
    While c a -> canBeWrong c || canCrash a
    a :-> b   -> any canCrash [a, b]
    x := e    -> canBeWrong e
  where
    canBeWrong (Duo Div _ _) = True
    canBeWrong (Duo Mod _ _) = True
    canBeWrong (Duo _ a b)   = any canBeWrong [a, b]
    canBeWrong (Uno _ e)     = canBeWrong e
    canBeWrong (Val _)       = False
    canBeWrong (Var _)       = False

crashed :: Trace a -> Bool
crashed Crash    = True
crashed End      = False
crashed Cut      = False
crashed (Step s) = crashed s
crashed (_ :> s) = crashed s

prop_crash =
  forAllShrink arbCommand typedShrink $ \p ->
  forAllShrink arbitrary shrink $ \(Positive n) ->
  let t = cut n (interp p) in
  whenFail (print t) $ crashed t ==> canCrash p

prop_typedExpr t =
  forAll genEnv $ \env ->
  forAllShrink (genExpr env t) typedShrinkE $ \e ->
    all ((t ==) . inferExpr env) $ typedShrinkE e

main = do
--  run 20 $ prop_typedExpr
  run 200 $ prop_Congruence
  run  50 $ prop_crash
--  run 100 $ prop_pretty_parse
--  run 2000 $ prop_Congruence'
    
run n = quickCheckWith stdArgs{ maxSuccess = n }

